%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%----------------------------------------------------------------------
%%% File    : erl_link_SUITE.erl
%%% Author  : Rickard Green <rickard.green@uab.ericsson.se>
%%% Purpose : Test erlang links
%%% Created : 13 Dec 2001 by Rickard Green <rickard.green@uab.ericsson.se>
%%%----------------------------------------------------------------------

-module(erl_link_SUITE).
-author('rickard.green@uab.ericsson.se').

%-define(line_trace, 1).
-include("test_server.hrl").

-export([all/1]).

% Test cases
-export([links/1,
	 dist_links/1,
	 monitor_nodes/1,
	 process_monitors/1,
	 dist_process_monitors/1]).

-export([init_per_testcase/2, fin_per_testcase/2]).

% Internal exports
-export([test_proc/0]).


-define(LNK_UNDEF, 0).
-define(LNK_LINK,  1). % normal link
-define(LNK_NODE,  2). % node link
-define(LNK_OMON,  3). % originating object monitor
-define(LNK_MON,   4). % terminating object monitor
-define(LNK_LINK1, 5). % uni-directional link, for monitor/2
-define(LNK_NODE1, 6). % node link, for monitor/2

-record(erl_link, {type = ?LNK_UNDEF,
		   item = [],
		   data = [],
		   ref  = []}).


all(suite) -> [links, dist_links, monitor_nodes, process_monitors,
	       dist_process_monitors].

links(doc) -> ["Tests node local links"];
links(suite) -> [];
links(Config) when list(Config) ->
    ?line common_link_test(node(), node()).

dist_links(doc) -> ["Tests distributed links"];
dist_links(suite) -> [];
dist_links(Config) when list(Config) ->
    ?line [NodeName] = get_names(1, dist_link),
    ?line {ok, Node} = start_node(NodeName),
    ?line common_link_test(node(), Node),
    ?line TP4 = spawn(?MODULE, test_proc, []),
    ?line TP5 = spawn(?MODULE, test_proc, []),
    ?line TP6 = spawn(Node, ?MODULE, test_proc, []),
    ?line true = tp_call(TP6, fun() -> link(TP4) end),
    ?line check_link(TP4, TP6),
    ?line true = tp_call(TP5,
			 fun() ->
				 process_flag(trap_exit,true),
				 link(TP6)
			 end),
    ?line check_link(TP5, TP6),
    ?line rpc:cast(Node, erlang, halt, []),
    ?line wait_until(fun () -> ?line is_proc_dead(TP4) end),
    ?line check_unlink(TP4, TP6),
    ?line true = tp_call(TP5,
			 fun() ->
				 receive
				     {'EXIT', TP6, noconnection} ->
					 true
				 end
			 end),
    ?line check_unlink(TP5, TP6),
    ?line tp_cast(TP5, fun() -> exit(normal) end),
    ?line ok.

common_link_test(NodeA, NodeB) ->
    ?line TP1 = spawn(NodeA, ?MODULE, test_proc, []),
    ?line check_unlink(TP1, self()),
    ?line TP2 = tp_call(TP1,
			fun () ->
				spawn_link(NodeB, ?MODULE, test_proc, [])
			end),
    ?line check_link(TP1, TP2),
    ?line true = tp_call(TP2, fun() -> unlink(TP1) end),
    ?line check_unlink(TP1, TP2),
    ?line true = tp_call(TP2, fun() -> link(TP1) end),
    ?line check_link(TP1, TP2),
    ?line false = tp_call(TP2, fun() -> process_flag(trap_exit, true) end),
    ?line tp_cast(TP1, fun () -> exit(died) end),
    ?line true = tp_call(TP2, fun() ->
				      receive 
					  {'EXIT', TP1, died} ->
					      true
				      end
			      end),
    ?line check_unlink(TP1, TP2),
    ?line TP3 = tp_call(TP2,
			fun () ->
				spawn_link(NodeA, ?MODULE, test_proc, [])
			end),
    ?line check_link(TP3, TP2),
    ?line tp_cast(TP2, fun() -> exit(died) end),
    ?line wait_until(fun () -> ?line is_proc_dead(TP3) end),
    ?line check_unlink(TP3, TP2),
    ?line ok.

monitor_nodes(doc) -> ["Tests monitor of nodes"];
monitor_nodes(suite) -> [];
monitor_nodes(Config) when list(Config) ->
    ?line [An, Bn, Cn, Dn] = get_names(4, dist_link),
    ?line {ok, A} = start_node(An),
    ?line {ok, B} = start_node(Bn),
    ?line C = list_to_atom(lists:concat([Cn, "@", hostname()])),
    ?line D = list_to_atom(lists:concat([Dn, "@", hostname()])),
    ?line 0 = no_of_monitor_node(self(), A),
    ?line 0 = no_of_monitor_node(self(), B),
    ?line monitor_node(A, true),
    ?line monitor_node(B, true),
    ?line monitor_node(D, true),
    ?line monitor_node(D, true),
    ?line monitor_node(A, false),
    ?line monitor_node(B, true),
    ?line monitor_node(C, true),
    ?line monitor_node(C, false),
    ?line monitor_node(C, true),
    ?line monitor_node(B, true),
    ?line monitor_node(A, false),
    ?line monitor_node(B, true),
    ?line monitor_node(B, false),
    ?line monitor_node(A, true),
    ?line check_monitor_node(self(), A, 1),
    ?line check_monitor_node(self(), B, 3),
    ?line check_monitor_node(self(), C, 0),
    ?line check_monitor_node(self(), D, 0),
    ?line receive {nodedown, C} -> ok end,
    ?line receive {nodedown, C} -> ok end,
    ?line receive {nodedown, C} -> ok end,
    ?line receive {nodedown, D} -> ok end,
    ?line receive {nodedown, D} -> ok end,
    ?line stop_node(A),
    ?line receive {nodedown, A} -> ok end,
    ?line check_monitor_node(self(), A, 0),
    ?line check_monitor_node(self(), B, 3),
    ?line stop_node(B),
    ?line receive {nodedown, B} -> ok end,
    ?line receive {nodedown, B} -> ok end,
    ?line receive {nodedown, B} -> ok end,
    ?line check_monitor_node(self(), B, 0),
    ?line receive
	      {nodedown, X} ->
		  ?line ?t:fail({unexpected_nodedown, X})
	  after 0 ->
		  ?line ok
	  end,
    ?line ok.
    

process_monitors(doc) -> ["Tests node local process monitors"];
process_monitors(suite) -> [];
process_monitors(Config) when list(Config) ->
    ?line common_process_monitors(node(), node()).

dist_process_monitors(doc) -> ["Tests distributed process monitors"];
dist_process_monitors(suite) -> [];
dist_process_monitors(Config) when list(Config) -> 
    ?line [Name] = get_names(1,dist_process_monitors),
    ?line {ok, Node} = start_node(Name),
    ?line common_process_monitors(node(), Node),
    ?line TP1 = spawn(Node, ?MODULE, test_proc, []),
    ?line R1 = erlang:monitor(process, TP1),
    ?line TP1O = get_down_object(TP1, self()),
    ?line check_process_monitor(self(), TP1, R1),
    ?line tp_cast(TP1, fun () -> halt() end),
    ?line receive
	      {'DOWN',R1,process,TP1O,noconnection} ->
		  ?line ok
	  end,
    ?line check_process_demonitor(self(), TP1, R1),
    ?line R2 = erlang:monitor(process, TP1),
    ?line receive
	      {'DOWN',R2,process,TP1O,noconnection} ->
		  ?line ok
	  end,
    ?line check_process_demonitor(self(), TP1, R2),
    ?line ok.


common_process_monitors(NodeA, NodeB) ->
    ?line TP1 = spawn(NodeA, ?MODULE, test_proc, []),
    ?line TP2 = spawn(NodeB, ?MODULE, test_proc, []),
    ?line run_common_process_monitors(TP1, TP2),
    ?line TP3 = spawn(NodeA, ?MODULE, test_proc, []),
    ?line TP4 = spawn(NodeB, ?MODULE, test_proc, []),
    ?line [TP4N] = get_names(1, common_process_monitors),
    ?line true = tp_call(TP4, fun () -> register(TP4N,self()) end),
    ?line run_common_process_monitors(TP3,
				      case node() == node(TP4) of
					  true -> TP4N;
					  false -> {TP4N, node(TP4)}
				      end),
    ?line ok.

run_common_process_monitors(TP1, TP2) ->
    ?line R1 = tp_call(TP1, fun () -> erlang:monitor(process, TP2) end),
    ?line check_process_monitor(TP1, TP2, R1),

    ?line tp_call(TP2, fun () -> catch erlang:demonitor(R1) end),
    ?line check_process_monitor(TP1, TP2, R1),

    ?line true = tp_call(TP1, fun () -> erlang:demonitor(R1) end),
    ?line check_process_demonitor(TP1, TP2, R1),

    ?line R2 = tp_call(TP1, fun () -> erlang:monitor(process, TP2) end),
    ?line TP2O = get_down_object(TP2, TP1),
    ?line check_process_monitor(TP1, TP2, R2),
    ?line tp_cast(TP2, fun () -> exit(bye) end),
    ?line wait_until(fun () -> ?line is_proc_dead(TP2) end),
    ?line ok = tp_call(TP1, fun () ->
				    ?line receive
					      {'DOWN',R2,process,TP2O,bye} ->
						  ?line ok
				    end
			    end),
    ?line check_process_demonitor(TP1, TP2, R2),
    
    ?line R3 = tp_call(TP1, fun () -> erlang:monitor(process, TP2) end),
    ?line ok = tp_call(TP1, fun () ->
				    ?line receive
					      {'DOWN',R3,process,TP2O,noproc} ->
						  ?line ok
				    end
			    end),
    ?line check_process_demonitor(TP1, TP2, R3),

    ?line tp_cast(TP1, fun () -> exit(normal) end),
    ?line wait_until(fun () -> ?line is_proc_dead(TP1) end),
    ?line ok.
    


%%
%% -- Internal utils --------------------------------------------------------
%%

% get_node(Name) when atom(Name) ->
%     ?line node();
% get_node({Name, Node}) when atom(Name) ->
%     ?line Node;
% get_node(NC) when pid(NC); port(NC); reference(NC) ->
%     ?line node(NC).

get_down_object(Item, _) when pid(Item) ->
    Item;
get_down_object({Name, Node} = Item, _) when atom(Name); atom(Node) ->
    Item;
get_down_object(Item, Watcher) when atom(Item), pid(Watcher) ->
    {Item, node(Watcher)};
get_down_object(Item, {_,Node}) when atom(Item), atom(Node) ->
    {Item, Node};
get_down_object(Item, Watcher) when atom(Item), atom(Watcher) ->
    {Item, node()}.

is_proc_dead(P) ->
    case is_proc_alive(P) of
	true -> false;
	false -> true
    end.

is_proc_alive(Pid) when pid(Pid), node(Pid) == node() ->
    ?line is_process_alive(Pid);
is_proc_alive(Name) when atom(Name) ->
    ?line case catch whereis(Name) of
	      Pid when pid(Pid) ->
		  ?line is_proc_alive(Pid);
	      _ ->
		  ?line false
	  end;
is_proc_alive({Name, Node}) when atom(Name), Node == node() ->
    ?line is_proc_alive(Name);
is_proc_alive(Proc) ->
    ?line is_remote_proc_alive(Proc).

is_remote_proc_alive({Name, Node}) when atom(Name), atom(Node) ->
    ?line is_remote_proc_alive(Name, Node);
is_remote_proc_alive(Pid) when pid(Pid) ->
    ?line is_remote_proc_alive(Pid, node(Pid));
is_remote_proc_alive(_) ->
    ?line false.

is_remote_proc_alive(PN, Node) ->		 
    ?line S = self(),
    ?line R = make_ref(),
    ?line monitor_node(Node, true),
    ?line P = spawn(Node, fun () -> S ! {R, is_proc_alive(PN)} end),
    ?line receive
	      {R, Bool} ->
		  ?line monitor_node(Node, false),
		  ?line Bool;
	      {nodedown, Node} ->
		  ?line false
	  end.

wait_until(Fun) ->
    ?line case Fun() of
	      true ->
		  ?line ok;
	      _ ->
		  ?line receive
			after 100 ->
				?line wait_until(Fun)
			end
	  end.

init_per_testcase(Func, Config) when atom(Func), list(Config) ->
    ?line Dog = ?t:timetrap(?t:minutes(1)),
    ?line [{watchdog, Dog}|Config].

fin_per_testcase(Func, Config) ->
    ?line Dog = ?config(watchdog, Config),
    ?line ?t:timetrap_cancel(Dog).


tp_call(Tp, Fun) ->
    ?line R = make_ref(),
    ?line Tp ! {call, self(), R, Fun},
    ?line receive
	      {R, Res} ->
		  ?line Res
	  end.

tp_cast(Tp, Fun) ->
    ?line Tp ! {cast, Fun}.

test_proc() ->
    ?line receive
	      {call, From, Ref, Fun} ->
		  ?line From ! {Ref, Fun()};
	      {cast, Fun} ->
		  ?line Fun()
	  end,
    ?line test_proc().


get_local_link_list(Obj) ->
    ?line case catch erlang:system_info({link_list, Obj}) of
	      LL when list(LL) ->
		  ?line LL;
	      _ ->
		  ?line []
	  end.

get_remote_link_list(Node, Obj) ->
    ?line case catch rpc:call(Node, erlang, system_info, [{link_list, Obj}]) of
	      LL when list(LL) ->
		  ?line LL;
	      _ ->
		  ?line []
	  end.


get_link_list({Node, DistEntry}) when Node == node(), atom(DistEntry) ->
    ?line get_local_link_list(DistEntry);
get_link_list({Node, DistEntry}) when atom(Node), atom(DistEntry) ->
    ?line get_remote_link_list(Node, DistEntry);
get_link_list(P) when pid(P); port(P) ->
    ?line case node(P) of
	      Node when Node == node() ->
		  ?line get_local_link_list(P);
	      Node ->
		  ?line get_remote_link_list(Node, P)
	  end;
get_link_list(undefined) ->
    ?line [].

find_erl_link(Obj, Ref) when reference(Ref) -> 
    ?line lists:foldl(fun (#erl_link{ref = R} = EL, Acc) when R == Ref ->
			      ?line [EL|Acc];
			  (_, Acc) ->
			      ?line Acc
		      end,
		      [],
		      get_link_list(Obj)).

find_erl_link(Obj, Type, [Item, Data]) when pid(Item);
					    port(Item);
					    atom(Item) -> 
    ?line lists:foldl(fun (#erl_link{type = T, item = I, data = D} = EL,
			   Acc) when T == Type, I == Item ->
			      ?line case Data of
					D ->
					    ?line [EL|Acc];
					[] ->
					    ?line [EL|Acc];
					_ ->
					    ?line Acc
				    end;
			  (_, Acc) ->
			      ?line Acc
		      end,
		      [],
		      get_link_list(Obj));
find_erl_link(Obj, Type, Item) when pid(Item); port(Item); atom(Item) ->
    find_erl_link(Obj, Type, [Item, []]). 

	

check_link(A, B) ->
    ?line [#erl_link{type = ?LNK_LINK,
		      item = B,
		      data = [],
		      ref = []}] = find_erl_link(A, ?LNK_LINK, B),
    ?line [#erl_link{type = ?LNK_LINK,
		      item = A,
		      data = [],
		      ref = []}] = find_erl_link(B, ?LNK_LINK, A),
    ?line case node(A) == node(B) of
	      false ->
		  ?line [#erl_link{type = ?LNK_LINK,
				   item = A,
				   data = B,
				   ref = []}] = find_erl_link({node(A),
							       node(B)},
							      ?LNK_LINK,
							      [A, B]),
		  ?line [#erl_link{type = ?LNK_LINK,
				   item = B,
				   data = A,
				   ref = []}] = find_erl_link({node(B),
							       node(A)},
							      ?LNK_LINK,
							      [B, A]);
	      true ->
		  ?line [] = find_erl_link({node(A), node(B)},
					   ?LNK_LINK,
					   [A, B]),
		  ?line [] = find_erl_link({node(B), node(A)},
					   ?LNK_LINK,
					   [B, A])
	  end,
    ?line ok.

check_unlink(A, B) ->
    ?line [] = find_erl_link(A, ?LNK_LINK, B),
    ?line [] = find_erl_link(B, ?LNK_LINK, A),
    ?line [] = find_erl_link({node(A), node(B)}, ?LNK_LINK, [A, B]),
    ?line [] = find_erl_link({node(B), node(A)}, ?LNK_LINK, [B, A]),
    ?line ok.

check_process_monitor(From, {Name, Node}, Ref) when pid(From),
					    atom(Name),
					    Node == node(From),
					    reference(Ref) ->
    ?line check_process_monitor(From, Name, Ref);
check_process_monitor(From, {Name, Node}, Ref) when pid(From),
					    atom(Name),
					    atom(Node),
					    reference(Ref) ->
    ?line MonitoredPid = rpc:call(Node, erlang, whereis, [Name]),
    ?line [#erl_link{type = ?LNK_LINK1,
		     item = From,
		     data = Node,
		     ref = Ref}] = find_erl_link(From, Ref),
    ?line [#erl_link{type = ?LNK_LINK1,
		     item = From,
		     data = Name,
		     ref = Ref}] = find_erl_link({node(From), Node}, Ref),
    ?line [#erl_link{type = ?LNK_LINK1,
		     item = From,
		     data = MonitoredPid,
		     ref = Ref}] = find_erl_link({Node, node(From)}, Ref),
    ?line [#erl_link{type = ?LNK_LINK1,
		     item = From,
		     data = Name,
		     ref = Ref}] = find_erl_link(MonitoredPid, Ref),
    ?line ok;
check_process_monitor(From, Name, Ref) when pid(From),
				    atom(Name),
				    undefined /= Name,
				    reference(Ref) ->
    ?line MonitoredPid = rpc:call(node(From), erlang, whereis, [Name]),
    ?line [#erl_link{type = ?LNK_LINK1,
		     item = From,
		     data = MonitoredPid,
		     ref = Ref}] = find_erl_link(From, Ref),
    ?line [#erl_link{type = ?LNK_LINK1,
		     item = From,
		     data = Name,
		     ref = Ref}] = find_erl_link(MonitoredPid, Ref);
check_process_monitor(From, To, Ref) when pid(From), pid(To), reference(Ref) ->
    ?line LL = [#erl_link{type = ?LNK_LINK1,
			  item = From,
			  data = To,
			  ref = Ref}],
    ?line LL = find_erl_link(From, Ref),
    ?line LL = find_erl_link(To, Ref),
    ?line case node(From) == node(To) of
	      false ->
		  ?line LL = find_erl_link({node(From), node(To)}, Ref),
		  ?line LL = find_erl_link({node(To), node(From)}, Ref);
	      true ->
		  ?line [] = find_erl_link({node(From), node(From)}, Ref)
	  end,
    ?line ok.


check_process_demonitor(From, {undefined, Node}, Ref) when pid(From),
							   reference(Ref) ->
    ?line [] = find_erl_link(From, Ref),
    ?line case node(From) == Node of
	      false ->
		  ?line [] = find_erl_link({node(From), Node}, Ref),
		  ?line [] = find_erl_link({Node, node(From)}, Ref);
	      true ->
		  ?line [] = find_erl_link({Node, Node}, Ref)
	  end,
    ?line ok;
check_process_demonitor(From, {Name, Node}, Ref) when pid(From),
						      atom(Name),
						      Node == node(From),
						      reference(Ref) ->
    ?line MonitoredPid = rpc:call(Node, erlang, whereis, [Name]),
    ?line case rpc:call(Node, erlang, whereis, [Name]) of
	      undefined ->
		  ?line check_process_demonitor(From, {undefined, Node}, Ref);
	      MonitoredPid ->
		  ?line check_process_demonitor(From, MonitoredPid, Ref)
	  end;
check_process_demonitor(From, {Name, Node}, Ref) when pid(From),
						      atom(Name),
						      atom(Node),
						      reference(Ref) ->
    ?line MonitoredPid = rpc:call(Node, erlang, whereis, [Name]),
    ?line [] = find_erl_link(From, Ref),
    ?line [] = find_erl_link({node(From), Node}, Ref),
    ?line [] = find_erl_link({Node, node(From)}, Ref),
    ?line [] = find_erl_link(MonitoredPid, Ref),
    ?line ok;
check_process_demonitor(From, undefined, Ref) when pid(From),
						   reference(Ref) ->
    ?line [] = find_erl_link(From, Ref),
    ?line case node(From) == node() of
	      false ->
		  ?line [] = find_erl_link({node(From), node()}, Ref),
		  ?line [] = find_erl_link({node(), node(From)}, Ref);
	      true ->
		  ?line [] = find_erl_link({node(), node()}, Ref)
	  end,
    ?line ok;
check_process_demonitor(From, Name, Ref) when pid(From),
					      atom(Name),
					      undefined /= Name,
					      reference(Ref) ->
    ?line check_process_demonitor(From, {Name, node()}, Ref);
check_process_demonitor(From, To, Ref) when pid(From),
					    pid(To),
					    reference(Ref) ->
    ?line [] = find_erl_link(From, Ref),
    ?line [] = find_erl_link(To, Ref),
    ?line case node(From) == node(To) of
	      false ->
		  ?line [] = find_erl_link({node(From), node(To)}, Ref),
		  ?line [] = find_erl_link({node(To), node(From)}, Ref);
	      true ->
		  ?line [] = find_erl_link({node(From), node(From)}, Ref)
	  end,
    ?line ok.

no_of_monitor_node(From, Node) when pid(From), atom(Node) ->
    ?line length(find_erl_link(From, ?LNK_NODE, Node)).

check_monitor_node(From, Node, No) when pid(From),
					atom(Node),
					integer(No),
					No >= 0 ->
    ?line LL = lists:duplicate(No, #erl_link{type = ?LNK_NODE, item = Node}),
    ?line DLL = lists:duplicate(No, #erl_link{type = ?LNK_NODE, item = From}),
    ?line LL = find_erl_link(From, ?LNK_NODE, Node),
    ?line DLL = find_erl_link({node(From), Node}, ?LNK_NODE, From),
    ?line ok.



hostname() ->
    ?line from($@, atom_to_list(node())).

from(H, [H | T]) -> T;
from(H, [_ | T]) -> from(H, T);
from(H, []) -> [].

get_names(N, T) when atom(T) ->
    get_names(N, T, []).
get_names(0, _, Acc) ->
    Acc;
get_names(N, T, Acc) ->
    {A, B, C} = now(),
    get_names(N-1, T, [list_to_atom(atom_to_list(?MODULE)
				    ++ "-"
				    ++ atom_to_list(T)
				    ++ "-"
				    ++ integer_to_list(A)
				    ++ "-"
				    ++ integer_to_list(B)
				    ++ "-"
				    ++ integer_to_list(C)) | Acc]).

start_node(Name) ->
    ?line start_node(Name, "").

start_node(Name, Args) ->
    ?line Pa = filename:dirname(code:which(?MODULE)),
    ?line ?t:start_node(Name, slave, [{args,  Args ++ " -pa " ++ Pa}]).
    

stop_node(Node) ->
    ?line ?t:stop_node(Node).

