%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(list_bif_SUITE).
-include("test_server.hrl").

-export([all/1,init_per_testcase/2,fin_per_testcase/2]).
-export([hd_test/1, tl_test/1, length/1, list_to_pid/1,
	 list_to_float/1]).


all(suite) ->
    [hd_test, tl_test, length, list_to_pid, list_to_float].

init_per_testcase(Case, Config) ->
    ?line Dog=test_server:timetrap(test_server:seconds(60)),
    [{watchdog, Dog}|Config].

fin_per_testcase(Case, Config) ->
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.


hd_test(suite) ->
    [];
hd_test(doc) ->
    ["Test hd/1 with correct and incorrect arguments"];
hd_test(Config) when list(Config) ->
    ?line 104=hd("hejsan"),
    ?line case catch hd(104) of
	      {'EXIT', {badarg, _}} -> ok;
	      Res ->
		  Str=io_lib:format("hd/1 with incorrect args "++
				   "succeeded.~nResult: ~p", [Res]),
		  test_server:fail(Str)
	  end,
    ok.


tl_test(suite) ->
    [];
tl_test(doc) ->
    ["Test tl/1 with correct and incorrect arguments"];
tl_test(Config) when list(Config) ->
    ?line "ejsan"=tl("hejsan"),
    ?line case catch tl(104) of
	      {'EXIT', {badarg, _}} ->
		  ok;
	      Res ->
		  Str=io_lib:format("tl/1 with incorrect args "++
				   "succeeded.~nResult: ~p", [Res]),
		  test_server:fail(Str)
	  end,
    ok.


length(suite) ->
    [];
length(doc) ->
    ["Test length/1 with correct and incorrect arguments"];
length(Config) when list(Config) ->
    ?line 6=length("hejsan"),
    ?line case catch length({tuple}) of
	      {'EXIT', {badarg, _}} ->
		  ok;
	      Res ->
		  Str=io_lib:format("length/1 with incorrect args "++
				   "succeeded.~nResult: ~p", [Res]),
		  test_server:fail(Str)
	  end,
    ok.
	      

list_to_pid(suite) ->
    [];
list_to_pid(doc) ->
    ["Test list_to_pid/1 with correct and incorrect arguments"];
list_to_pid(Config) when list(Config) ->
    ?line Me=self(),
    ?line MyListedPid=pid_to_list(Me),
    ?line Me=list_to_pid(MyListedPid),
    ?line case catch list_to_pid("Incorrect list") of
	      {'EXIT', {badarg, _}} ->
		  ok;
	      Res ->
		  Str=io_lib:format("list_to_pid/1 with incorrect "++
				   "arg succeeded.~nResult: ~p",
				    [Res]),
		  test_server:fail(Str)
	  end,
    ok.


list_to_float(suite) ->
    [];
list_to_float(doc) ->
    ["Test list_to_float/1 with correct and incorrect arguments"];
list_to_float(Config) when list(Config) ->
    ?line 5.89000=list_to_float("5.89"),
    ?line 5.89898=list_to_float("5.89898"),
    ?line case catch list_to_float("58") of
	      {'EXIT', {badarg, _}} -> ok;
	      Res ->
		  Str=io_lib:format("list_to_float with incorrect "++
				   "arg succeeded.~nResult: ~p",
				    [Res]),
		  test_server:fail(Str)
	  end,
    ok.

